/* 
   Passifist
   Copyright (C) Patrik Karlsson <patrik@cqure.net> 2003
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 *
 * Plugin for detecting broadcasting tftp requests
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "net.h"
#include "hsrp.h"
#include "util.h"
#include "plugin.h"

struct host *pHost = NULL;

void plugin_dbstorehost(struct hsrphost *);

void printhost(FILE *pFD, char *prefix, struct hsrphost *host) {
  unsigned char mac[20];

  mactostr((unsigned char *)&mac, host->ether);
  fprintf(pFD, "%s,%s,%s\n", prefix, mac, host->ip );

}

int dupcheck(struct hsrphost *list, struct hsrphost *host) {

  struct hsrphost *loop = list;

  while ( loop ) {
    
    if ( strcmp( loop->ip, host->ip ) == 0 && 
	 strcmp( loop->state, host->state ) == 0 )
      return 1;
    
    loop = loop->next;

  }

  return 0;

}

/*
 * Frees memory of all hosts in linked list
 *
 */
void freehosts( struct hsrphost *hosts ) {

  struct hsrphost *curr, *rmv;

  curr = hosts;

  while ( curr ) {
    rmv = curr;
    curr = curr->next;
    free(rmv);
  }

}

struct hsrphost *pushost(struct hsrphost *hosts, struct hsrphost *newhost) {

  if ( dupcheck( hosts, newhost ) == 0x00 ) {
    plugin_dbstorehost( newhost );
    newhost->next = hosts;
    hosts = newhost;
  }

  return hosts;

}

struct hsrphost *createhost(char *ether, char *ip, char state, char *authdata,
			    struct in_addr virtip ){

  struct hsrphost *newhost;

  newhost = (struct hsrphost *) malloc ( sizeof( struct hsrphost ) );
  memcpy( newhost->ether, ether, sizeof( newhost->ether ) );
  memcpy( newhost->ip, ip, sizeof( newhost->ip ) );
  
  bzero( newhost->state, sizeof( newhost->state ) );
  bzero( newhost->pw, sizeof( newhost->pw ) );
  bzero( newhost->virtip, sizeof( newhost->virtip ) );

  if ( state == 0x10 )
    strcpy( newhost->state, "Active");
  else
    strcpy( newhost->state, "Standby");


  memcpy( newhost->pw, authdata, sizeof( newhost->pw ) );
  strncpy( newhost->virtip, inet_ntoa( virtip ), sizeof( newhost->virtip ) - 1 );

  return newhost;

}

int plugin_manglepacket( const u_char *pPacket, int nLen, int nVerbose ) {

  int nOffset;
  struct hsrphost *newhost;
  struct in_addr *inetaddr;
  struct ip_header *pIpheader;
  struct udphdr *pUdpheader;
  struct ether_header *pEthheader;
  struct hsrp_packet *pHsrp;

  u_int8_t *pSrcEth;

  nOffset = sizeof( struct ether_header ) + sizeof( struct ip_header );

  if ( nLen < nOffset ) {
    return -1;
  }

  pEthheader = (struct ether_header *)pPacket;
  pSrcEth = pEthheader->ether_shost;

  pIpheader = (struct ip_header *)(pPacket + nOffset - sizeof( struct ip_header ) );
  inetaddr = (struct in_addr *)&pIpheader->ip_src;

  if ( pIpheader->ip_p != IPPROTO_UDP ) {
    return -1;
  }

  pUdpheader = (struct udphdr *)(pPacket + sizeof(struct ether_header) +
				 IP_HL(pIpheader) * 4 );

  nOffset += sizeof( struct udphdr );

  if ( nLen < nOffset ) {
    return -1;
  }

  if ( htons(pUdpheader->uh_dport) == 1985 && htons( pUdpheader->uh_sport) == 1985 ) {
    if ( nLen < nOffset + sizeof( struct hsrp_packet ) )
      return -1;

    pHsrp = (struct hsrp_packet *)( pPacket + nOffset );

    newhost = createhost( pSrcEth, inet_ntoa( *inetaddr  ), pHsrp->state, 
			  pHsrp->authdata, pHsrp->virtip);
    hsrphosts = pushost( hsrphosts, newhost );

    return 0;
  }

  return -1;

}

void plugin_report(FILE *pFD) {

  struct hsrphost *loop = hsrphosts;

  while ( loop ) {
    printhost(pFD, "cim", loop);
    loop = loop->next;
  }

}

static int gethostcount() {
  
  struct hsrphost* loop = hsrphosts;
  int nHosts = 0;

  while ( loop ) {
    nHosts ++;
    loop = loop->next;
  }

  return nHosts;

}

void plugin_stats(FILE *pFD) {

}

void plugin_stop() {
  freehosts(hsrphosts);
}

int plugin_packetcount() {
  return nPacketCount;
}

void plugin_dbstorehost(struct hsrphost *h) {

  unsigned char mac[20];

  mactostr((unsigned char *)&mac, h->ether);

  pHost = plug_storehost( pHost, mac, h->ip, "", "hsrp");

  if ( !pHost )
    return;

  nPacketCount ++;

  plug_storeproperty( pHost, "state", h->state );
  plug_storeproperty( pHost, "authdata", h->pw );
  plug_storeproperty( pHost, "virtip", h->virtip );

}

struct host *plugin_gethost() {
  return pHost;
}

void plugin_freehost() {
  plug_freehost( pHost );
  pHost = NULL;
}

char *plugin_getfilter() {
  return "udp && port 1985";
}

void *plugin_getdesc() {
  return "HSRP - Cisco Hotstandby Routing Protocol plugin";
}
