/* 
   Passifist
   Copyright (C) Patrik Karlsson <patrik@cqure.net> 2003
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 *
 * Plugin for detecting Netop Management SW
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "net.h"
#include "netop.h"
#include "util.h"
#include "plugin.h"

void plugin_dbstorehost(struct netophost *);

struct host *pHost = NULL;
struct property *pProp = NULL;


void printhost(FILE *pFD, char *prefix, struct netophost *host) {
  unsigned char mac[20];

  mactostr((unsigned char *)&mac, host->ether);
  fprintf(pFD, "%s,%s,%s,%s\n", prefix, mac, host->ip, host->hostname );
}

int dupcheck(struct netophost *list, struct netophost *host) {

  struct netophost *loop = list;

  while ( loop ) {
    
    if ( strcmp( loop->ip, host->ip ) == 0 )
      return 1;
    
    loop = loop->next;

  }

  return 0;

}

/*
 * Frees memory of all hosts in linked list
 *
 */
void freehosts( struct netophost *hosts ) {

  struct netophost *curr, *rmv;

  curr = hosts;

  while ( curr ) {
    rmv = curr;
    curr = curr->next;
    free(rmv);
  }

}

struct netophost *pushost(struct netophost *hosts, struct netophost *newhost) {

  if ( dupcheck( hosts, newhost ) == 0x00 ) {
    newhost->next = hosts;
    hosts = newhost;
    plugin_dbstorehost( newhost );
  }

  return hosts;

}

/* 
   netop hostnames are padded with spaces....
   so lets do some heavy truncating ....
*/
void truncname( char *name ) {

  int i;
  
  for ( i=0; i<32; i++ ) {
    if ( name[i] == 0x20 ) {
      name[i] = 0;
      return;
    }
  }

}

struct netophost *createhost(char *ether, char *ip, char *pName) {

  struct netophost *newhost;

  newhost = (struct netophost *) malloc ( sizeof( struct netophost ) );
  memcpy( newhost->ether, ether, sizeof( newhost->ether ) );
  memcpy( newhost->ip, ip, sizeof( newhost->ip ) );
  memset( newhost->hostname, 0, sizeof( newhost->hostname ) );
  memcpy( newhost->hostname, pName, sizeof( newhost->hostname ) - 1 );
  truncname( newhost->hostname );

  return newhost;

}

int plugin_manglepacket( const u_char *pPacket, int nLen, int nVerbose ) {

  int nOffset;
  struct netophost *newhost;
  struct in_addr *inetaddr;
  struct ip_header *pIpheader;
  struct udphdr *pUdpheader;
  struct ether_header *pEthheader;
  u_int8_t *pSrcEth;
  u_int8_t *pHostname;

  nOffset = sizeof( struct ether_header ) + sizeof( struct ip_header );
  
  if ( nLen < nOffset ) {
    return -1;
  }

  pEthheader = (struct ether_header *)pPacket;
  pSrcEth = pEthheader->ether_shost;

  pIpheader = (struct ip_header *)(pPacket + nOffset - sizeof( struct ip_header ) );
  inetaddr = (struct in_addr *)&pIpheader->ip_src;

  if ( pIpheader->ip_p != IPPROTO_UDP ) {
    return -1;
  }

  pUdpheader = (struct udphdr *)(pPacket + sizeof(struct ether_header) +
				 IP_HL(pIpheader) * 4 );

  nOffset += sizeof( struct udphdr );

  if ( nLen < nOffset ) {
    return -1;
  }

  if ( htons(pUdpheader->uh_sport) == 6502 ) {

    /* offset to host-name */
    nOffset += 30;

    /* offset + hostname length */
    if ( nLen < nOffset + 32 )
      return -1;
      
    pHostname = (char *)(pPacket + nOffset );

    newhost = createhost( pSrcEth, inet_ntoa( *inetaddr  ), pHostname );
    if ( nVerbose )
      printhost( stderr, "netop", newhost );
    netophosts = pushost( netophosts, newhost );
    nPacketCount ++;

    return 0;
  }

  return -1;
  
}

void plugin_report(FILE *pFD) {

  struct netophost *loop = netophosts;

  while ( loop ) {
    printhost(pFD, "netop", loop);
    loop = loop->next;
  }

}


static int gethostcount() {
  
  struct netophost* loop = netophosts;
  int nHosts = 0;

  while ( loop ) {
    nHosts ++;
    loop = loop->next;
  }

  return nHosts;

}

void plugin_stats(FILE *pFD) {
  fprintf( pFD, "\n");
  fprintf( pFD, "---- NetOp Plugin Statistics ------------------------\n");
  fprintf( pFD, "Netop Hosts: %d\n", gethostcount());
  fprintf( pFD, "-----------------------------------------------------\n");
}


void plugin_stop() {
  freehosts(netophosts);
}

int plugin_packetcount() {
  return nPacketCount;
}

void plugin_dbstorehost(struct netophost *h) {

  unsigned char mac[20];

  mactostr((unsigned char *)&mac, h->ether);
  pHost = plug_storehost( pHost, mac, h->ip, h->hostname, "netop");

}


struct host *plugin_gethost() {
  return pHost;
}

struct property *plugin_getproperty() {
  return pProp;
}

void plugin_freehost() {
  plug_freehost( pHost );
  pHost = NULL;
}

void plugin_freeproperty() {
  plug_freeproperty( pProp );
  pProp = NULL;
}

char *plugin_getdesc() {
  return "NETOP - Netop remote control software plugin";
}
