/* 
   Passifist
   Copyright (C) Patrik Karlsson <patrik@cqure.net> 2003
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "smb.h"
#include "util.h"
#include "plugin.h"
#include "net.h"

struct host *pHost = NULL;
struct property *pProp = NULL;

void plugin_dbstorehost(struct smbhost *);
void plugin_dbstoredom(struct smbdomain *);

/*
 * Checks host against list for duplicates
 *
 */
int dupcheck(struct smbhost *list, const u_char md5[16] ) {

  struct smbhost *loop = list;

  while ( loop ) {

    if ( md5comp( loop->md5, md5 ) == 0 )
      return 1;

    loop = loop->next;
  }

  return 0;

}

int dupcheckdom(struct smbdomain *list, struct smbdomain *dom) {

  struct smbdomain *loop = list;

  while ( loop ) {

    if ( strcmp( loop->domain, dom->domain ) == 0 )
      return 1;

    loop = loop->next;
  }

  return 0;

}

/*
 * Frees memory of all hosts in linked list
 *
 */
void freehosts( struct smbhost *hosts ) {

  struct smbhost *curr, *rmv;

  curr = hosts;

  while ( curr ) {
    rmv = curr;
    curr = curr->next;
    free(rmv);
  }

}

void freedoms( struct smbdomain *doms ) {

  struct smbdomain *curr, *rmv;

  curr = doms;

  while ( curr ) {
    rmv = curr;
    curr = curr->next;
    free(rmv);
  }

}

struct smbdomain *pushdom(struct smbdomain *domains, struct smbdomain *new ) {

  if ( dupcheckdom( domains, new ) == 0 ) {
    new->next = domains;
    domains = new;
  }
  
  return domains;

}

struct smbdomain *createdom(const char *ether, const struct in_addr ip,
			    const char *domain, const char *mb) {

  struct smbdomain *dom;

  dom = (struct smbdomain *) malloc ( sizeof( struct smbdomain ) );

  if ( !dom )
    return NULL;

  memcpy( dom->ether, ether, sizeof( dom->ether ) );
  dom->ip = ip;
  strncpy( dom->domain, domain, sizeof( dom->domain ) );
  strncpy( dom->masterbrowser, mb, sizeof( dom->masterbrowser ) );

  plugin_dbstoredom( dom );

  return dom;

}

/*
 * creates a new host structure from data
 *
 */
struct smbhost *createhost(const u_int16_t proto,
			   const char *ether, const struct in_addr ip,
			   const char *nbname, const u_char osmajor,
			   const u_char osminor, const long srvtype,
			   const char *comment) {

  struct smbhost *host;

  u_char *md5dig;
  char data[128];
  int nLen;

  nLen = snprintf( data, sizeof( data ), "%d,%s,%s,%lu", proto, 
		   nbname, inet_ntoa(ip), srvtype);

  //fprintf(stdout, "%s\n", data);

  data[nLen] = '\0';

  md5dig = createMD5( data, nLen );

  if ( dupcheck( smbhosts, md5dig ) > 0 ) {
    free( md5dig );
    return NULL;
  }

  host = (struct smbhost *)malloc( sizeof( struct smbhost ) );

  if ( !host ) {
    free( md5dig );
    return NULL;
  }

  memcpy( host->md5, md5dig, sizeof( host->md5 ) );
  free( md5dig );

  memcpy( host->ether, ether, sizeof( host->ether ) );
  host->ip = ip;
  memset( host->nbname, 0, sizeof( nbname ) );
  memcpy( host->nbname, nbname, sizeof(host->nbname) );
  host->osmajor = osmajor;
  host->osminor = osminor;
  host->srvtype = srvtype;
  host->proto = proto;
  memset( host->comment, 0, sizeof( host->comment ) );
  strncpy(host->comment, comment, sizeof(host->comment) - 1);

  return host;
}

/*
 * Pushes a new host into the linked list
 *
 */
struct smbhost *pushhost(struct smbhost *hosts, struct smbhost *newhost ) {

  plugin_dbstorehost( newhost );
  newhost->next = hosts;
  hosts = newhost;

  return hosts;

}

/*
 * Decodes netbios name and returns type
 */
int decodenbname( const u_char *pIn, char *pNBname ) {

  int nLen = (*pIn++) / 2;
  int nType = 0;

  if ( nLen > 30 || nLen < 1 ) {
    return -1;
  }

  while ( nLen-- ) {

    if (pIn[0] < 'A' || pIn[0] > 'P' || pIn[1] < 'A' || pIn[1] > 'P') {
      *pNBname = 0;
      return -1;
    }

    *pNBname = ((pIn[0] - 'A') << 4) + (pIn[1] - 'A');
    pIn += 2;
    pNBname++;

  } 

  *pNBname = 0;
  nType = pNBname[-1];

  return nType;

}

void dissectnb( const u_char *pUdpPacket, u_int8_t *pSrcEth, int nLen ) {
  
  struct nb_ns_hdr *nbnshdr;
  struct nb_ns_query *nbnsquery;
  struct nb_ns_record *nbnsrecord;
  struct smbhost *newhost;

  int nOffset, nQuestions;
  short nFlags;
  char nbname[32];
  char nNBType;
  
  nOffset = sizeof( struct nb_ns_hdr );

  if ( nLen < nOffset ) {
    fprintf(stderr, "NB Packet to small\n");
    return;
  }
  
  nbnshdr = ( struct nb_ns_hdr * )( pUdpPacket + nOffset - sizeof( struct nb_ns_hdr ));
  
  nFlags = (htons(nbnshdr->flags) << 1 ) & 0xf000;

  /* if this is not a registration return */
  if ( nFlags != 0x5000 )
    return;

  nQuestions = htons(nbnshdr->questions);
  nOffset += sizeof( struct nb_ns_query );

  if ( nLen < nOffset ) {
    fprintf(stderr, "NB Packet to small\n");
    return;
  }
  
  nbnsquery = ( struct nb_ns_query *)( pUdpPacket + nOffset - sizeof ( struct nb_ns_query ));
  nNBType = decodenbname( nbnsquery->name, (char *)&nbname );  

  nOffset += sizeof( struct nb_ns_record );
  nbnsrecord = ( struct nb_ns_record *)( pUdpPacket + nOffset - 
					 sizeof ( struct nb_ns_record ));

  if ( nNBType == 0x00 && ( htons(nbnsrecord->flags) != 0x8000) ) {
    newhost = createhost(0x0800, pSrcEth, nbnsrecord->address, nbname, 0, 0, 0, "");

    if ( newhost )
      smbhosts = pushhost( smbhosts, newhost );

    if ( nVerbose )
      printhost( stderr, "nb", newhost );
  }
    


}

void dissectbrowserproto( const u_char *pPacket, u_int8_t *pSrcEth, int nLen ) {

  char nbsrcname[32];
  char nbdstname[32];
  char *pComment;
  char *pMBrowser;
  int nSrcNBType;
  int nDstNBType;
  int nOffset;
  u_int16_t nProto = 0x0800; /*ip*/
  struct in_addr ip;

  struct nb_datagram_svc *nbdgrsvc;
  struct netbios *nb;
  struct smb_header *smbhdr;
  struct smb_mailslot *mailslot;
  struct ms_browseproto_ha *msbrowsha;
  struct ms_browseproto_da *msbrowsda;
  struct smbhost *newhost;
  struct smbdomain *newdomain;

  u_char *pSlotName;

  /* direct group datagram */
  if ( pPacket[0] == 0x11 ) {

    nOffset = sizeof( struct nb_datagram_svc );

    if ( nLen < nOffset )
      return;

    nbdgrsvc = (struct nb_datagram_svc *)(pPacket + nOffset - sizeof( struct nb_datagram_svc ));

    nSrcNBType = decodenbname( nbdgrsvc->srcname, (char *)&nbsrcname );
    nDstNBType = decodenbname( nbdgrsvc->dstname, (char *)&nbdstname );

    nbdgrsvc->sport = htons(nbdgrsvc->sport);
    nbdgrsvc->len = htons(nbdgrsvc->len);
    nbdgrsvc->offset = htons(nbdgrsvc->offset );
    ip = nbdgrsvc->sip;
  }
  /* process netbeui traffic */
  else {

    nOffset = sizeof( struct netbios );

    if ( nLen < nOffset )
      return;

    nb = ( struct netbios *)(pPacket);
    memset( nbsrcname, 0, sizeof( nbsrcname ) );
    memset( nbdstname, 0, sizeof( nbdstname ) );
    memcpy( nbsrcname, nb->sendname, sizeof( nb->sendname ) - 1 );
    memcpy( nbdstname, nb->recvname, sizeof( nb->recvname ) - 1);

    nSrcNBType = nb->sendname[16];
    nDstNBType = nb->recvname[16];

    nProto = 0x0000; /* unset = netbeui */
    ip = inet_makeaddr(0, 0);
  }

  smbhdr = (struct smb_header *)( pPacket + nOffset );

  nOffset += sizeof( struct smb_header );
  nOffset += sizeof( struct trans_req );
  nOffset += sizeof( struct smb_mailslot );

  if ( nLen < nOffset )
    return;

  mailslot = (struct smb_mailslot *)(pPacket + nOffset - sizeof( struct smb_mailslot ) );

  /* the remaining data should be our mailslot */
  if ( nLen - nOffset != mailslot->size )
    return;

  pSlotName = ( u_char * ) (pPacket + nOffset);

  if ( strcmp( pSlotName, "\\MAILSLOT\\BROWSE" ) != 0 )
    return;

  nOffset += strlen( pSlotName ) + 1;

  if ( nLen < nOffset )
    return;

  /* host announcement or local master announcement */
  if ( *(pPacket + nOffset) == 1 || *(pPacket + nOffset) == 0x0f ) {
    nOffset += sizeof( struct ms_browseproto_ha);

    pComment = (char *)( pPacket + nOffset );

    if ( nLen < nOffset + strlen(pComment) ) {
      fprintf(stderr, "Corrupted packet\n");
      return;
    }
    
    nOffset -= sizeof( struct ms_browseproto_ha);

    msbrowsha = ( struct ms_browseproto_ha *)(pPacket + nOffset );

    newhost = createhost(nProto, pSrcEth, ip, 
			 msbrowsha->hostname, msbrowsha->osmajor,
			 msbrowsha->osminor, msbrowsha->servertype,
			 pComment);

    if ( newhost )
      smbhosts = pushhost(smbhosts, newhost);

  }
  /* domain announcement ? */
  else if ( *(pPacket + nOffset) == 0x0c ) {
    nOffset += sizeof( struct ms_browseproto_da);

    pMBrowser = (char *)(pPacket + nOffset);

    if ( nLen < nOffset + strlen( pMBrowser ) ) {
      fprintf(stderr, "ERROR: Launch the missiles and call the president\n");
      return;
    }

    nOffset -= sizeof( struct ms_browseproto_da);
    msbrowsda = ( struct ms_browseproto_da *)(pPacket + nOffset );
    newdomain = createdom(pSrcEth, ip, msbrowsda->domain, pMBrowser );

    smbdoms = pushdom(smbdoms, newdomain);

    if ( nVerbose )
      printdomain( stderr, "smbdom", newdomain );

  }

}

/*
 * Mangles the udp packet
 *
 */
int mangleudp( const u_char *pPacket, int nLen ) {

  struct ether_header *pEther;
  struct ip_header *pIpheader;
  struct udphdr *pUdpheader;
  short nUdpLen;
  u_char *pUdpData;
  int nTotSize;
  u_int8_t *pSrcEth;

  nTotSize = (sizeof( struct ip_header) + sizeof( struct ether_header ) +
	      sizeof( struct udphdr ) );

  if ( nLen < nTotSize ) {
    fprintf(stderr, "ERROR: mangleudp() packet too small\n");
    return -1;
  }

  pEther = (struct ether_header *) pPacket;
  pSrcEth = pEther->ether_shost;
  pIpheader = (struct ip_header *)(pPacket + sizeof(struct ether_header));
  pUdpheader = (struct udphdr *)(pPacket + sizeof(struct ether_header) +
				 IP_HL(pIpheader) * 4 );

  nTotSize += ntohs(pUdpheader->uh_ulen) - sizeof( struct udphdr );

  if ( nLen < nTotSize ) {
    fprintf(stderr, "ERROR: mangleudp() udppacket too small\n");
    fprintf(stderr, "DEBUG: nTotSize=%d, nLen=%d\n", nTotSize, nLen );
    return -1;
  }
  
  nUdpLen = ntohs(pUdpheader->uh_ulen);

  /* hmm.... */
  if ( nUdpLen > 1500 ) {
    fprintf(stderr, "ERROR: OOOPS UDP packet is to big!\n");
    return -1;
  }

  pUdpData = (u_char *)pUdpheader + sizeof( struct udphdr );

  if ( htons(pUdpheader->uh_dport) == 138 )
    dissectbrowserproto( (u_char *)pUdpData, pSrcEth,
			 nUdpLen - sizeof( struct udphdr ) );
  else if ( htons(pUdpheader->uh_dport) == 137 ) {
    dissectnb( (u_char *)pUdpData, pSrcEth,
	       nUdpLen - sizeof( struct udphdr ) );
  }

  return 0;

}

int plugin_manglepacket( const u_char *pPacket, int nLen, int nVerb ) {

  struct ether_header *pEther;
  struct ip_header *pIpheader;
  u_int8_t *pLLC;
  u_int8_t *pSrcEth;

  nVerbose = nVerb;

  if ( nLen < sizeof( struct ether_header ) )
    return -1;

  pEther = (struct ether_header *) pPacket;
  pSrcEth = pEther->ether_shost;

  /* check to see if we have an ip packet */
  if (ntohs (pEther->ether_type) == ETHERTYPE_IP) {

    if ( nLen < (sizeof( struct ip_header) + sizeof( struct ether_header )) ) {
      return -1;
    }

    pIpheader = (struct ip_header*)(pPacket + sizeof(struct ether_header));
    if ( pIpheader->ip_p != IPPROTO_UDP ) {
      return -1;
    }
    else {
      //fprintf(stderr, "DEBUG: Mangling UDP packet\n");
      if ( mangleudp( pPacket, nLen ) == 0 )
	return 0;
    }

  } /* end ip packet */
  /* 
     Do some hardcore netbeui parsing
  */
  else {
    
    if ( nLen < sizeof( struct ether_header ) + 3 )
      return -1;

    pLLC = ( u_int8_t *)(pPacket + sizeof( struct ether_header )); 
    /* LLC is Netbios */
    if ( pLLC[0] == 0xF0 && pLLC[1] == 0xF0 ) {
      pPacket = ( u_int8_t *)(pPacket + sizeof( struct ether_header ) + 3); 
      dissectbrowserproto( pPacket , pSrcEth, 
			   nLen - sizeof( struct ether_header ) - 3);
    }

  }

  return -1;

}


void printdomain(FILE *pFD, char *prefix, struct smbdomain *dom) {

  fprintf(pFD, "%s,%s,%s\n", prefix, dom->domain, dom->masterbrowser);

}

void printhost(FILE *pFD, char *prefix, struct smbhost *host) {

  unsigned char mac[20];
  char proto[8];
  char ip[16];

  memset( proto, 0, sizeof(proto));
  if ( host->proto == 0x00 )
    strcpy( proto, "netbeui");
  else
    strcpy( proto, "ip");

  mactostr((unsigned char *)&mac, host->ether);

  memset(ip, 0, sizeof( ip ) );
  strncpy(ip, inet_ntoa( host->ip ), sizeof( ip ) - 1 );

  if ( host->ip.s_addr == 0 )
    strcpy(ip, "N/A");

  fprintf(pFD, "%s,%s,%s,%s", prefix, mac, ip, host->nbname);
    
  if ( host->osmajor == 0x05 )
    fprintf(pFD, ",W2K" );
  else if ( host->osmajor == 0x04 )
    fprintf(pFD, ",NT4" );
  else if ( host->osmajor == 0x03 && host->osminor == 0x33 )
    fprintf(pFD, ",NT3.51");
  
  if ( host->srvtype & SV_TYPE_SERVER )
    fprintf(pFD, ",Srv");
  if ( host->srvtype & SV_TYPE_WORKSTATION )
    fprintf(pFD, ",Wks");
  if ( host->srvtype & SV_TYPE_SQLSERVER )
    fprintf(pFD, ",Sql");
  if ( host->srvtype & SV_TYPE_DOMAIN_CTRL )
    fprintf(pFD, ",PDC");
  if ( host->srvtype & SV_TYPE_DOMAIN_BAKCTRL)
    fprintf(pFD, ",BDC");
  if ( host->srvtype & SV_TYPE_TIME_SOURCE )
    fprintf(pFD, ",Time");
  if ( host->srvtype & SV_TYPE_AFP )
    fprintf(pFD, ",AFP");
  if ( host->srvtype & SV_TYPE_NOVELL )
    fprintf(pFD, ",Novell");
  if ( host->srvtype & SV_TYPE_DOMAIN_MEMBER )
    fprintf(pFD, ",Dom. Member");
  if ( host->srvtype & SV_TYPE_PRINTQ_SERVER )
    fprintf(pFD, ",Print");
  if ( host->srvtype & SV_TYPE_DIALIN_SERVER )
    fprintf(pFD, ",RAS");
  if ( host->srvtype & SV_TYPE_MASTER_BROWSER )
    fprintf(pFD, ",Master browser");
  if ( host->srvtype & SV_TYPE_TERMINALSERVER )
    fprintf(pFD, ",Term. Srv");
  
    
  //  fprintf(stdout, ",%x", host->srvtype);
  
  fprintf(pFD, "\n");

}

void outputdomains(FILE *pFD, struct smbdomain *doms) {
  
  struct smbdomain *loop = doms;

  while ( loop ) {
    printdomain(pFD, "smbdom", loop );
    loop = loop->next;
  }

}

void outputhosts(FILE *pFD, struct smbhost *hosts) {

  struct smbhost *loop = hosts;

  while ( loop ) {
    printhost(pFD, "smbhost", loop);
    loop = loop->next;
  }

}

int getnetbeuihostcount() {

  struct smbhost *loop = smbhosts;
  int nCount = 0;

  while ( loop ) {
    if ( loop->proto == 0 )
      nCount ++;
    loop = loop->next;
  }

  return nCount;

}

int gethostcount() {

  struct smbhost *loop = smbhosts;
  int nCount = 0;

  while ( loop ) {
    nCount ++;
    loop = loop->next;
  }

  return nCount;

}

int getdomaincount() {

  struct smbdomain *loop = smbdoms;
  int nCount = 0;

  while ( loop ) {
    nCount ++;
    loop = loop->next;
  }

  return nCount;

}

int gethostsoftype( int nType ) {

  struct smbhost *loop = smbhosts;
  int nCount = 0;

  while ( loop ) {
    if ( loop->srvtype & nType )
      nCount ++;

    loop = loop->next;
  }

  return nCount;

}

void plugin_stats(FILE *pFD) {
  fprintf( pFD, "\n");
  fprintf( pFD, "---- SMB Plugin Statistics --------------------------\n");
  fprintf( pFD, "Windows hosts: %d\n", gethostcount());
  fprintf( pFD, "Windows domains/workgroups: %d\n", getdomaincount());
  fprintf( pFD, "SQL Servers: %d\n", gethostsoftype(SV_TYPE_SQLSERVER));
  fprintf( pFD, "Terminal Servers: %d\n", gethostsoftype(SV_TYPE_TERMINALSERVER));
  fprintf( pFD, "Host running netbeui: %d\n", getnetbeuihostcount());
  fprintf( pFD, "-----------------------------------------------------\n");
}

void plugin_report(FILE *pFD) {
  outputhosts(pFD, smbhosts);
  outputdomains(pFD, smbdoms);
}

char *plugin_getfilter() {
  return pMask;
}

int plugin_packetcount() {
  return nPacketCount;
}

void plugin_stop() {
  freehosts( smbhosts );
  freedoms( smbdoms );
}

char *getservices( char *services, int nLen, long srvtype ) {

  char srvc[1024];
  
  memset( srvc, 0, sizeof( srvc ) );

  if ( srvtype & SV_TYPE_SERVER )
    strcat(srvc, ",Srv");
  if ( srvtype & SV_TYPE_WORKSTATION )
    strcat(srvc, ",Wks");
  if ( srvtype & SV_TYPE_SQLSERVER )
    strcat(srvc, ",Sql");
  if ( srvtype & SV_TYPE_DOMAIN_CTRL )
    strcat(srvc, ",PDC");
  if ( srvtype & SV_TYPE_DOMAIN_BAKCTRL)
    strcat(srvc, ",BDC");
  if ( srvtype & SV_TYPE_TIME_SOURCE )
    strcat(srvc, ",Time");
  if ( srvtype & SV_TYPE_AFP )
    strcat(srvc, ",AFP");
  if ( srvtype & SV_TYPE_NOVELL )
    strcat(srvc, ",Novell");
  if ( srvtype & SV_TYPE_DOMAIN_MEMBER )
    strcat(srvc, ",Dom. Member");
  if ( srvtype & SV_TYPE_PRINTQ_SERVER )
    strcat(srvc, ",Print");
  if ( srvtype & SV_TYPE_DIALIN_SERVER )
    strcat(srvc, ",RAS");
  if ( srvtype & SV_TYPE_MASTER_BROWSER )
    strcat(srvc, ",Master browser");
  if ( srvtype & SV_TYPE_TERMINALSERVER )
    strcat(srvc, ",Term. Srv");



  
  /* avoid first comma */
  strncpy( services, srvc + 1, nLen - 1 );

  return services;

}

char *getversion( int osmajor, int osminor ) {

  if ( osmajor == 0x03 && osminor == 51 )
    return "Windows NT 3.51";
  else if ( osmajor == 0x04 && osminor == 0x00 )
    return "Windows NT 4.0";
  else if ( osmajor == 0x05 && osminor == 0x00 )
    return "Windows 2000";

  return NULL;

}

void plugin_dbstoredom(struct smbdomain *d) {

  unsigned char mac[20];

  mactostr((unsigned char *)&mac, d->ether);

  pHost = plug_storehost( pHost, mac, inet_ntoa(d->ip), 
			  d->masterbrowser, "smbhost" );

  if ( pHost ) {
    nPacketCount ++;
    plug_storeproperty( pHost, "srvtype", "Master browser");
    plug_storeproperty( pHost, "domain", d->domain );
  }


}

void plugin_dbstorehost(struct smbhost *h) {

  unsigned char mac[20];
  unsigned char val[20];
  char *pOSVersion;
  char services[1024];

  memset( services, 0, sizeof( services ) );
  mactostr((unsigned char *)&mac, h->ether);
  pOSVersion = NULL;

  /* this is most probably a host announcmement */
  if ( h->srvtype > 0 ) {

    if ( h->osmajor > 0 ) {
      pOSVersion = getversion( h->osmajor, h->osminor );
    }

    if ( h->srvtype > 0 ) {
      getservices((char *)&services, sizeof( services ), h->srvtype);
    }

    pHost = plug_storehost( pHost, mac, inet_ntoa(h->ip), h->nbname, "smbhost");
    if ( !pHost )
      return;

    nPacketCount ++;

  }
  else {
    pHost = plug_storehost( pHost, mac, inet_ntoa(h->ip),
				 h->nbname, "nbhost");
    
    if ( !pHost )
      return;

    nPacketCount ++;
  }
  
  if ( pOSVersion )
    plug_storeproperty( pHost, "osver", pOSVersion);
  else if ( h->osmajor > 0 ) {
    sprintf((char *)&val, "%d%d", h->osmajor, h->osminor);
    plug_storeproperty( pHost, "osver", val);
  }
  
  if ( strlen( h->comment ) )
    plug_storeproperty( pHost, "comment", h->comment);
  
  if ( h->srvtype > 0 )
    plug_storeproperty( pHost, "srvtype", services);
  

}

struct host *plugin_gethost() {
  return pHost;
}

struct property *plugin_getproperty() {
  return pProp;
}

void plugin_freehost() {
  plug_freehost( pHost );
  pHost = NULL;
}

void plugin_freeproperty() {
  plug_freeproperty( pProp );
  pProp = NULL;
}

char *plugin_getdesc() {
  return "SMB - Microsoft SMB/Netbios plugin";
}
